
Program VarRecs;

(* Démonstration de l'utilisation de fichier d'enregistrements de taille variable.
   Dans cet exemple, il s'agit d'un fichier de citations contenant trois chaînes
   de caractères et un entier. *)


Uses
  Classes,
  SysUtils,
  Crt,
  Strings;


Const
  NomFichier = 'citations.dat';   (* Nom du fichier *)


Type
  (* Elément d'une liste chaînée de citations *)
  PCitation = ^TCitation;
  TCitation = record
    Index : Integer;
    (* Données de la citation *)
    Texte : pChar;
    Opus : pChar;
    Page : Word;
    Clefs : pChar;
    (* Liens avec la liste chaînée *)
    pPrec, pSuiv : PCitation;
  end;


Const
  (* Tête et queue de la liste chaînée des citations *)
  TeteListe : PCitation = Nil;
  QueueListe : PCitation = Nil;


Var
  (* Action choisie dans le menu principal *)
  Action : Char;


Procedure AjoutElementListeChainee (const Citation : PCitation; var TeteListe, QueueListe : PCitation);
(* Ajout d'un nouvel élément dans la liste chaînée *)
Begin
  if TeteListe = Nil
     then
       TeteListe := Citation
     else
       begin
         QueueListe^.pSuiv := Citation;
         Citation^.pPrec := QueueListe;
         Citation^.Index := QueueListe^.Index + 1;
       end;
  QueueListe := Citation;
End;


Procedure AjoutCitation (var TeteListe, QueueListe : PCitation);
(* Lecture des données d'une citation et ajout de celle-ci à la liste chaînée *)
Var
  NouvelleCitation : PCitation;      (* Nouvel élément à ajouter *)
  Buffer : Array [0..999] of Char;   (* Buffer de lecture de chaîne *)
  Erreur : Integer;                  (* Code d'erreur renvoyé par Val *)
Begin
  try
    (* Création d'un nouvel élément zéroté *)
    New(NouvelleCitation);
    FillChar(NouvelleCitation^,SizeOf(NouvelleCitation^),0);
    NouvelleCitation^.Index := 1;
    (* lecture des données *)
    WriteLn('Entrez les données de la nouvelle citation :');
    Write('   Texte : ');
    ReadLn(Buffer);
    NouvelleCitation^.Texte := StrNew(Buffer);
    Write('   Opus  : ');
    ReadLn(Buffer);
    NouvelleCitation^.Opus := StrNew(Buffer);
    Write('   Page  : ');
    ReadLn(Buffer);
    Val(Buffer,NouvelleCitation^.Page,Erreur);
    Write('   Clefs : ');
    ReadLn(Buffer);
    NouvelleCitation^.Clefs := StrNew(Buffer);
    (* Ajout du nouvel élément dans la liste chaînée *)
    AjoutElementListeChainee(NouvelleCitation,TeteListe,QueueListe);
  except
    WriteLn('*** Erreur d''allocation de mémoire ! ***');
  end;
End;


Procedure AffichageCitation (const Citation : PCitation);
(* Affichage des données d'une citation *)
Begin
  with Citation^ do
    begin
      WriteLn('"',Texte,'"');
      WriteLn('   Opus  : ',Opus);
      WriteLn('   Page  : ',Page);
      WriteLn('   Clefs : ',Clefs);
    end;
End;


Procedure AffichageListeCitations (const TeteListe, QueueListe : PCitation);
(* Navigation dans la liste des citations *)
Var
  Citation : PCitation;
  Action : Char;
Begin
  Citation := TeteListe;
  if Citation = Nil
     then
       begin
         WriteLn('*** La liste est vide ! ***');
         Exit;
       end;
  repeat
    if Citation <> Nil
       then
         AffichageCitation(Citation);
    (* Lecture de l'action désirée par l'utilisateur *)
    Write('=> (P)récédente - (S)uivante - (D)ébut - (F)in - (Q)uitter : ');
    Action := ReadKey;
    WriteLn;
    case Action of
      'D','d' : Citation := TeteListe;
      'F','f' : Citation := QueueListe;
      'P','p' : if Citation <> Nil
                   then
                     Citation := Citation^.pPrec;
      'S','s' : if Citation <> Nil
                   then
                     Citation := Citation^.pSuiv;
    end;
  until Action in ['Q','q'];
End;


Procedure DesallocationCitations (var TeteListe, QueueListe : PCitation);
(* Désallocation de la liste chaînée des citations *)
Var
  pCit1, pCit2 : PCitation;
Begin
  pCit1 := TeteListe;
  while (pCit1 <> Nil) do
    begin
      pCit2 := pCit1^.pSuiv;
      (* Désallocation des chaînes dynamiques *)
      with pCit1^ do
        begin
          StrDispose(Texte);
          StrDispose(Opus);
          StrDispose(Clefs);
        end;
      (* Désallocation de l'élément lui-même *)
      Dispose(pCit1);
      pCit1 := pCit2;
    end;
  TeteListe := Nil;
  QueueListe := Nil;
End;


Procedure ChargementCitations (var TeteListe, QueueListe : PCitation);
(* Chargement des citations depuis un fichier d'enregistrements de taille variable *)
Var
  NouvelleCitation : PCitation;
  Fichier : TFileStream;
  Buffer : Array [0..999] of Char;   (* Buffer de lecture de chaîne *)
  Compteur : LongInt;                (* Compteur d'enregistrements *)
  NbEnregistrements : LongInt;       (* Nombre d'enregistrements dans le fichier *)
  Taille : LongInt;                  (* Taille du champ qui suit *)
Begin
  try
    Fichier := TFileStream.Create(NomFichier,fmOpenRead);
    (* lecture du nombre d'enregistrements *)
    Fichier.Read(NbEnregistrements,SizeOf(NbEnregistrements));
    (* lecture des citations *)
    Compteur := 1;
    while (Compteur <= NbEnregistrements) do
      begin
        (* Création d'un nouvel élément zéroté *)
        New(NouvelleCitation);
        FillChar(NouvelleCitation^,SizeOf(NouvelleCitation^),0);
        NouvelleCitation^.Index := 1;
        (* Remplissage des champs *)
        Fichier.Read(Taille,SizeOf(Taille));
        Fichier.Read(Buffer,Taille);
        NouvelleCitation^.Texte := StrNew(Buffer);
        Fichier.Read(Taille,SizeOf(Taille));
        Fichier.Read(Buffer,Taille);
        NouvelleCitation^.Opus := StrNew(Buffer);
        Fichier.Read(NouvelleCitation^.Page,SizeOf(NouvelleCitation^.Page));
        Fichier.Read(Taille,SizeOf(Taille));
        Fichier.Read(Buffer,Taille);
        NouvelleCitation^.Clefs := StrNew(Buffer);
        (* Ajout du nouvel élément à la liste chaînée *)
        AjoutElementListeChainee(NouvelleCitation,TeteListe,QueueListe);
        Inc(Compteur);
      end;
    Fichier.Free;
  except
    WriteLn('*** Erreur de lecture du fichier ! ***');
  end;
End;


Procedure SauvegardeCitations (var TeteListe, QueueListe : PCitation);
(* Sauvegarde des citations dans un fichier d'enregistrements de taille variable *)
Var
  Citation : PCitation;
  Fichier : TFileStream;
  Taille : LongInt;   (* Taille du champ qui suit *)
Begin
  Citation := TeteListe;
  try
    Fichier := TFileStream.Create(NomFichier,fmCreate);
    (* Sauvegarde du nombre d'enregistrements *)
    Fichier.Write(QueueListe^.Index,SizeOf(QueueListe^.Index));
    (* Sauvegarde des citations *)
    while (Citation <> Nil) do
      begin
        with Citation^ do
          begin
            Taille := StrLen(Texte);
            Fichier.Write(Taille,SizeOf(Taille));
            Fichier.Write(Texte[0],Taille);
            Taille := StrLen(Opus);
            Fichier.Write(Taille,SizeOf(Taille));
            Fichier.Write(Opus[0],Taille);
            Fichier.Write(Page,SizeOf(Page));
            Taille := StrLen(Clefs);
            Fichier.Write(Taille,SizeOf(Taille));
            Fichier.Write(Clefs[0],Taille);
          end;
        Citation := Citation^.pSuiv;
      end;
    Fichier.Free;
  except
    WriteLn('*** Erreur lors de la sauvegarde du fichier ! ***');
  end;
End;

(* ----------- Programme principal ---------- *)

Begin
  (* Chargement automatique des citations au début du programme *)
  ChargementCitations(TeteListe,QueueListe);
  repeat
    WriteLn;
    WriteLn('---------------------------------------');
    WriteLn(' Recueil de citations - Menu principal');
    WriteLn('---------------------------------------');
    WriteLn('(N)ouvelle citation');
    WriteLn('(A)fficher les citations');
    WriteLn('(Q)uitter');
    Write('=> Votre choix : ');
    Action := ReadKey;
    WriteLn;
    case Action of
      'N','n' : AjoutCitation(TeteListe,QueueListe);
      'A','a' : AffichageListeCitations(TeteListe,QueueListe);
    end;
  until Action in ['Q','q'];
  (* Sauvegarde automatique du fichier *)
  if TeteListe <> Nil
     then
       SauvegardeCitations(TeteListe,QueueListe);
  WriteLn('Réalisé par Alcatîz pour Developpez.com');
  (* Désallocation des variables dynamiques *)
  DesallocationCitations(TeteListe,QueueListe);
End.

